/*
 * funcs.h
 *
 * rcsid = "/net/bcomp/src/master/nexus/funcs.h,v 1.54 1995/11/13 22:51:40 tuecke Exp"
 */

#ifndef _NEXUS_INCLUDE_FUNCS_H
#define _NEXUS_INCLUDE_FUNCS_H

/*
 * args.c
 */
extern void	_nx_get_args(int *argc,
			     char ***argv,
			     char *args_env_variable,
			     char *package_designator,
			     void (*usage_message_func)(void),
			     int (*new_process_params_func)(char *, int),
			     nexus_bool_t ignore_command_line_args);
extern void	_nx_args_init(int *argc, char ***argv);
extern void	_nx_args_cleanup(int *argc, char ***argv);
extern int	_nx_new_process_params(char *buf, int size);
extern int	_nx_split_args(char *arg_string,
			       int starting_argc,
			       int max_argc,
			       char **package_argv,
			       char *string_buffer);
extern char *	_nx_get_package_id_start(void);
extern char *	_nx_get_package_id_end(void);
extern char *	_nx_get_argv0(void);


/*
 * db_iface.c
 */
extern void	_nx_database_usage_message(void);
extern int	_nx_database_new_process_params(char *buf, int size);
extern void	_nx_database_init(int *argc, char ***argv,
				  nexus_module_list_t module_list[]);
extern void	_nx_database_shutdown(void);
extern void	_nx_database_abort(void);
extern char *   _nx_database_hash_table_lookup(char *name, 
					       int number,
					       char *key);
extern void     _nx_database_hash_table_add(char *name,
				            int number,
				            int count,
				            nexus_list_t *attr);
extern nexus_list_t *_nx_database_parse_attributes(char *buf,
					           int buf_len,
					           char *starting_point,
					           FILE *fp);
extern void     _nx_database_flush_rest_of_line(char *buf,
					        int buf_len,
					        char *starting_point,
					        FILE *fp);
extern void     _nx_database_hash_table_add_nodes_with_attrs(
			nexus_db_hash_entry_t *nodes,
			nexus_list_t *attrs);


/*
 * context.c
 */
extern void	_nx_context_usage_message(void);
extern void	_nx_context_init(int *argc, char ***argv);
extern nexus_context_t *_nx_context_alloc(int initial_segment_size,
					  nexus_bool_t is_node);
extern int	_nx_context_new_process_params(char *buf, int size);
extern void	_nx_context_register_handlers(void);
extern void	_nx_context_checkin(nexus_global_pointer_t *context_gp,
				    int nexus_boot_rc,
				    nexus_global_pointer_t *reply_gp,
				    int checkin_number
#ifdef NEXUS_PROFILE
				    , int need_master_gp
#endif			 
				    );
#ifdef NEXUS_PROFILE
extern void _nx_wait_for_master_gp_from_creator(nexus_context_t *context);
extern void _nx_send_master_gp_to_context(nexus_global_pointer_t *context_gp);
#endif


/*
 * fault.c
 */
extern void	_nx_fault_tolerance_usage_message(void);
extern int	_nx_fault_tolerance_new_process_params(char *buf, int size);
extern void	_nx_fault_tolerance_init(int *argc, char ***argv);
extern int	_nx_fault_detected(int fault_code);


/*
 * handler.c
 */
extern void	_nx_handler_usage_message(void);
extern int	_nx_handler_new_process_params(char *buf, int size);
extern void	_nx_handler_init(int *argc, char ***argv);
extern void	_nx_handler_init_for_context(nexus_context_t *context);
extern void	_nx_handler_destroy(nexus_context_t *context);
extern void	_nx_lookup_handler(unsigned long context,
				   char *handler_name,
				   int handler_id,
				   nexus_handler_type_t *type,
				   nexus_handler_func_t *func);
extern void	_nx_handle_message(char *handler_name,
				   int handler_id,
				   nexus_handler_type_t type,
				   nexus_handler_func_t func,
				   unsigned long context_arg,
				   unsigned long address,
#ifdef NEXUS_PROFILE
				   int node_id,
				   int context_id,
				   int message_length,
#endif
				   void *buffer);
extern void	_nx_dump_rsr_profile_info(nexus_context_t *context);
#ifdef NEXUS_PROFILE
extern void	_nx_accumulate_rsr_profile_info(nexus_context_t *context,
						nexus_handler_record_t *rec,
						int node_id,
						int context_id,
						int message_length);
#endif /* NEXUS_PROFILE */


/*
 * init.c
 */
extern void	_nx_init_nexus(int *argc,
			       char ***argv,
			       char *args_env_variable,
			       char *package_designator,
			       int (*package_args_init_func)(int *argc,char ***argv),
			       void (*usage_message_func)(void),
			       int (*new_process_params_func)(char *,int),
			       nexus_module_list_t module_list[],
			       nexus_node_t **nodes,
			       int *n_nodes);
extern void	_nx_exit_transient_process(int rc);


/*
 * nl_*.c
 */
extern void	_nx_nodelock_register_handlers(void);
extern void	_nx_nodelock_master_init(void);
extern nexus_bool_t	_nx_nodelock_check(void);
extern void	_nx_nodelock_shutdown(void);
extern void	_nx_nodelock_cleanup(void);


/*
 * md_*.c
 */
extern void	_nx_md_usage_message(void);
extern void	_nx_md_init(int *argc, char ***argv);
extern void	_nx_md_shutdown(void);
extern void	_nx_md_abort(int return_code);
extern void	_nx_md_exit(int return_code);
extern int	_nx_md_get_command_line_args(int starting_argc,
					     int max_argc,
					     char **argv);
extern char *	_nx_md_system_error_string(int the_errno);
extern char *	_nx_md_get_unique_session_string();
extern void	_nx_md_gethostname(char *name, int len);
extern int	_nx_md_getpid();
extern int	_nx_md_new_process_params(char *buf, int size);
extern int	_nx_md_fork(void);
extern int	_nx_md_block_sigchld(void);
extern void	_nx_md_deliver_sigchld(int mask);
extern void	_nx_md_wait_for_children(void);


/*
 * pr_iface.c
 */
extern void	_nx_proto_usage_message(void);
extern int	_nx_proto_new_process_params(char *buf, int size);
extern void	_nx_proto_init(int *argc, char ***argv,
			       nexus_module_list_t module_list[]);
extern void	_nx_proto_shutdown(nexus_bool_t shutdown_others);
extern void	_nx_proto_abort(void);
extern nexus_bool_t	_nx_must_poll(void);
extern void	_nx_proto_get_creator_gp_params(char *buf,
						int buf_size,
						nexus_global_pointer_t *gp);
extern void	_nx_proto_construct_creator_gp(
					 nexus_global_pointer_t *creator_gp);
extern int	_nx_write_global_pointer(int filedesc,
					 nexus_global_pointer_t *gp);
extern int	_nx_read_global_pointer(int filedesc,
					nexus_global_pointer_t *gp);
extern int	_nx_write(int fd, char *buf, int len);
extern int	_nx_read(int fd, char *buf, int len);
extern int	_nx_write_int4(int fd, int i);
extern int	_nx_read_int4(int fd, int *i);


/*
 * pr_local.c
 */
extern void *	_nx_pr_local_info(void);


/*
 * process.c
 */
extern void	_nx_process_usage_message(void);
extern void	_nx_process_init(int *argc,
				 char ***argv,
				 char *package_designator,
				 int (*package_args_init_func)(int *argc,
							       char **argv[]),
				 nexus_module_list_t module_list[],
				 nexus_node_t **nodes,
				 int *n_nodes);
extern void	_nx_process_start(void);
extern int	_nx_process_new_process_params(char *buf, int size);
extern void	_nx_process_register_handlers(void);
#ifdef NEXUS_PROFILE
extern void	_nx_get_next_node_id_from_master(int *node_id,
						 int n_nodes,
						 nexus_context_t *context);
#endif /* NEXUS_PROFILE */

/*
 * st_iface.c
 */
extern void	_nx_startup_usage_message(void);
extern int	_nx_startup_new_process_params(char *buf, int size);
extern void     _nx_startup_modify_command_for_type(nexus_startup_node_t *node,
					    	    char *cmd,
					    	    int max_cmd_length,
						    char ***environment,
						    int *n_env);
extern void	_nx_startup_register_handlers(void);
extern nexus_bool_t	_nx_startup_preinit(nexus_module_list_t module_list[]);
extern void	_nx_startup_init(int *argc, char ***argv);
extern char *	_nx_startup_get_master_node_name(void);
extern void	_nx_startup_initial_nodes(nexus_node_t *my_node,
					  nexus_node_t **nodes,
					  int *n_nodes);
extern int	_nx_startup_context(char *executable_path,
				    int context_size,
				    nexus_global_pointer_t *reply_gp,
				    int checkin_number,
				    nexus_context_t **new_local_context);
extern void	_nx_startup_shutdown(nexus_bool_t shutdown_others);
extern void	_nx_startup_abort(void);
extern void	_nx_node_checkin(nexus_node_t *node,
				 int n_nodes,
				 nexus_global_pointer_t *creator_gp,
				 int checkin_number
#ifdef NEXUS_PROFILE
				 , int need_master_gp
#endif			 
				 );


/*
 * th_*.c
 */
#ifndef NEXUS_LITE
extern void	_nx_thread_usage_message(void);
extern int	_nx_thread_new_process_params(char *buf, int size);
extern void	_nx_thread_preinit(void);
extern void	_nx_thread_init(int *argc, char ***argv);
extern void	_nx_thread_shutdown(void);
extern void	_nx_thread_abort(int return_code);
extern void	_nx_thread_create_handler_thread(void);
extern void	_nx_thread_shutdown_handler_thread(void);
#endif /* NEXUS_LITE */


/*
 * trace.c
 */
extern void	_nx_traceback(void);


/*
 * util.c
 */
extern char *	_nx_copy_string(char *s);
extern char *	_nx_executable_name(void);
extern void	_nx_strip_tmp_mnt_from_path(char *dir);
extern char *   _nx_find_attribute(char *attr,
				   char *search_string,
				   char separator);
extern void     _nx_get_next_value(char *values,
				   char separator,
				   char **next,
				   char **this);
extern void	_nx_hex_encode_byte_array(unsigned char *bytes,
					  int length,
					  char *hex);
extern void	_nx_hex_decode_byte_array(char *hex,
					  int length,
					  unsigned char *bytes);
#ifdef NEXUS_NO_STRTOUL
extern unsigned long strtoul(const char *nptr,
			     char **endptr,
			     int base);
#endif /* NEXUS_NO_STRTOUL */

#endif /* _NEXUS_INCLUDE_FUNCS_H */

