/*
 * macros.h
 *
 * rcsid = "/net/bcomp/src/master/nexus/macros.h,v 1.19 1995/11/13 22:51:42 tuecke Exp"
 */

#ifndef _NEXUS_INCLUDE_MACROS_H
#define _NEXUS_INCLUDE_MACROS_H

/*
 * _nx_context()
 *
 * Get the context for the current thread and save it
 * into *Context (nexus_context_t **).
 */
#ifdef NEXUS_LITE
#define _nx_context(Context) \
    *(Context) = _nx_lite_context
#else
#define _nx_context(Context) \
    *(Context) = (nexus_context_t *)nexus_thread_getspecific(_nx_context_key)
#endif


/*
 * _nx_set_context()
 *
 * Set Context to be the context of the current thread.
 */
#ifdef NEXUS_LITE
#define _nx_set_context(Context) \
    _nx_lite_context = (Context)
#else
#define _nx_set_context(Context) \
    nexus_thread_setspecific(_nx_context_key, (void *) (Context));
#endif


#ifndef NEXUS_LITE

/*
 * _nx_thread_context()
 *
 * Set *Context (nexus_context_t **) to be the context for the passed thread.
 */
#define _nx_thread_context(Thread, Context) \
    *(Context) = ((Thread)->context)

/*
 * _nx_thread_set_context()
 *
 * Set the context (nexus_context_t *) for the passed thread.
 *
 * This may be used by a message handling thread to set itself
 * to the appropriate context before handling a rsr.
 */
#define _nx_thread_set_context(Thread, Context) \
    (Thread)->context = (Context)

#endif /* NEXUS_LITE */


/*
 * _nx_context_id()
 *
 * Set *Context_ID (int *) to be the context id for the context
 * of the calling thread.
 */
#define _nx_context_id(Context_ID) \
{ \
    nexus_context_t *__context; \
    _nx_context(&__context); \
    if (__context) \
        *(Context_ID) = __context->id; \
    else \
        *(Context_ID) = -1; \
}


/*
 * _nx_node_id()
 *
 * Set *Node_ID (int *) to be the node id for the node
 * of the calling thread.
 */
#define _nx_node_id(Node_ID) \
{ \
    *(Node_ID) = _nx_my_node_id; \
}


/*
 * PackLiba(char *Liba, unsigned long Context, unsigned long Address)
 */
#define PackLiba(Liba, Context, Address) \
{ \
    unsigned long __l[2]; \
    __l[0] = (Context); \
    __l[1] = (Address); \
    memcpy((Liba), __l, (2*sizeof(unsigned long))); \
}


/*
 * UnpackLiba(nexus_byte_t *Liba,
 *            unsigned long Context, unsigned long Address)
 */
#define UnpackLiba(Liba, Context, Address) \
{ \
    unsigned long __l[2]; \
    memcpy(__l, (Liba), (2*sizeof(unsigned long))); \
    (Context) = __l[0]; \
    (Address) = __l[1]; \
}


/*
 * UnpackLibaFromGP(nexus_global_pointer_t *GP,
 *                  unsigned long Context, unsigned long Address)
 */
#define UnpackLibaFromGP(GP, Context, Address) \
{ \
    unsigned long __l[2]; \
    if (gp->liba_is_inline) \
    { \
        memcpy(__l, ((GP)->liba.array), (2*sizeof(unsigned long))); \
    } \
    else \
    { \
        memcpy(__l, ((GP)->liba.pointer), (2*sizeof(unsigned long))); \
    } \
    (Context) = __l[0]; \
    (Address) = __l[1]; \
}


/*
 * PackInt1(nexus_byte_t *Array, int Index, int Integer)
 */
#define PackInt1(Array, Index, Integer) \
{ \
    (Array)[(Index)++] = (nexus_byte_t)  ((Integer) & 0xFF); \
}

/*
 * UnpackInt1(nexus_byte_t *Array, int Index, int Integer)
 */
#define UnpackInt1(Array, Index, Integer) \
{ \
    (Integer) = (int) (Array)[(Index)]; \
    (Index) += 1; \
}


/*
 * PackInt2(nexus_byte_t *Array, int Index, int Integer)
 */
#define PackInt2(Array, Index, Integer) \
{ \
    (Array)[(Index)++] = (nexus_byte_t) (((Integer) & 0xFF00) >> 8); \
    (Array)[(Index)++] = (nexus_byte_t)  ((Integer) & 0xFF); \
}

/*
 * UnpackInt2(nexus_byte_t *Array, int Index, int Integer)
 */
#define UnpackInt2(Array, Index, Integer) \
{ \
    (Integer) = (  ( ((int) (Array)[(Index)]) << 8) \
		 |   ((int) (Array)[(Index)+1]) ); \
    (Index) += 2; \
}


/*
 * PackInt4(nexus_byte_t *Array, int Index, int Integer)
 */
#define PackInt4(Array, Index, Integer) \
{ \
    (Array)[(Index)++] = (nexus_byte_t) (((Integer) & 0xFF000000) >> 24); \
    (Array)[(Index)++] = (nexus_byte_t) (((Integer) & 0xFF0000) >> 16); \
    (Array)[(Index)++] = (nexus_byte_t) (((Integer) & 0xFF00) >> 8); \
    (Array)[(Index)++] = (nexus_byte_t)  ((Integer) & 0xFF); \
}

/*
 * UnpackInt4(nexus_byte_t *Array, int Index, int Integer)
 */
#if NEXUS_INT_SIZE == 4
#define UnpackInt4(Array, Index, Integer) \
{ \
    (Integer) = (  ( ((int) (Array)[(Index)])   << 24) \
		 | ( ((int) (Array)[(Index)+1]) << 16) \
		 | ( ((int) (Array)[(Index)+2]) << 8) \
		 |   ((int) (Array)[(Index)+3]) ); \
    (Index) += 4; \
}
#else
#define UnpackInt4(Array, Index, Integer) \
{ \
    (Integer) = (  ( ((int) (Array)[(Index)])   << 24) \
		 | ( ((int) (Array)[(Index)+1]) << 16) \
		 | ( ((int) (Array)[(Index)+2]) << 8) \
		 |   ((int) (Array)[(Index)+3]) ); \
    if ((Integer) & 0x80000000) /* Sign extend */ \
    { \
	(Integer) |= 0xFFFFFFFF00000000; \
    } \
    (Index) += 4; \
}
#endif


/*
 * PackMIProtoEntry(nexus_byte_t *Array, int Index,
 *                  nexus_proto_type_t Type, int Size, nexus_byte_t *SubArray)
 */
#define PackMIProtoEntry(Array, Index, Type, Size, SubArray) \
{ \
    PackInt2(Array, Index, (int) (Type)); \
    PackInt2(Array, Index, Size); \
    memcpy(&((Array)[(Index)]), SubArray, Size) ; \
    (Index) += (Size); \
}


/*
 * UnpackMIProtoEntry(nexus_byte_t *Array, int Index,
 *               nexus_proto_type_t Type, int Size, nexus_byte_t *SubArray)
 */
#define UnpackMIProtoEntry(Array, Index, Type, Size, SubArray) \
{ \
    int __i; \
    UnpackInt2(Array, Index, __i); \
    (Type) = (nexus_proto_type_t) __i; \
    UnpackInt2(Array, Index, Size); \
    (SubArray) = &((Array)[(Index)]); \
    (Index) += (Size); \
}


/*
 * UnpackMIProtoHeader(nexus_byte_t *Array, int Index,
 *                     int Node_number, int Context_differentiator,
 *                     char *Node_name)
 */
#define UnpackMIProtoHeader(Array, Index, Node_number, Context_differentiator, Node_name) \
{ \
    UnpackInt4(Array, Index, Node_number); \
    UnpackInt4(Array, Index, Context_differentiator); \
    (Node_name) = (char *) &((Array)[(Index)]); \
    (Index) += (strlen(Node_name) + 1); \
}


#ifdef NEXUS_DEBUG
nexus_bool_t NexusDbgState( unsigned long, unsigned long,
			      unsigned long, unsigned long );
#define NexusDebug(Level) (_nx_debug_level >= (Level) || NexusDbgState( 0xffffffff, 0xffffffff, 0xffffffff, Level ))
#endif

#ifdef NEXUS_DEBUG
#define nexus_debug_printf(level, message) \
do { \
    if (NexusDebug(level)) \
    { \
	nexus_printf message; \
    } \
} while (0)
#define nexus_new_debug_printf(catagory, module, operation, level, message) \
do { \
    if (NexusDbgState(catagory, module, operation, level)) \
    { \
	nexus_printf message; \
    } \
} while (0)
#else
#define nexus_debug_printf(level, message)
#define nexus_new_debug_printf(level, message)
#endif

#ifdef NEXUS_DEBUG
#define NexusBufferMagicCheck(func, buf) \
{ \
    if (*(buf) == (nexus_buffer_t) NULL) \
	nexus_fatal(#func ": NULL nexus_buffer_t\n"); \
    if ((*(buf))->magic != NEXUS_BUFFER_MAGIC) \
	nexus_fatal(#func ": Bad magic number %x on buffer %x\n", \
		    (*(buf))->magic, buf); \
}
#else
#define NexusBufferMagicCheck(func, buf)
#endif


#ifdef NEXUS_LITE
/*
 * We need to spoof a few functions in the NexusLite version.
 */
#define _nx_thread_id(Thread_ID) *(Thread_ID) = 0;
#endif /* NEXUS_LITE */

#endif /* _NEXUS_INCLUDE_MACROS_H */
