/*
 * nexus.h
 *
 * rcsid = "/net/bcomp/src/master/nexus/nexus.h,v 1.74 1995/11/13 22:51:44 tuecke Exp"
 *
 * This header contains the exported interface of Nexus.
 */

#ifndef _NEXUS_INCLUDE_NEXUS_H
#define _NEXUS_INCLUDE_NEXUS_H

/* EXTERN_C_BEGIN and EXTERN_C_END should surround any Nexus prototypes in
   nexus.h or Nexus .h files that are included by nexus.h.  This will
   allow C++ codes to include nexus.h and properly link against the
   Nexus library.
*/
#ifdef __cplusplus
#define EXTERN_C_BEGIN extern "C" {
#define EXTERN_C_END }
#else
#define EXTERN_C_BEGIN
#define EXTERN_C_END
#endif

EXTERN_C_BEGIN

/*
 * By defining this, Nexus will always try to use macros unless
 * this is overridden by NEXUS_DO_NOT_USE_MACROS.
 */
#define NEXUS_USE_MACROS


/*
 * nexus_path_type_t
 */
typedef enum _nexus_path_type_t
{
    NEXUS_PATH_TYPE_DATABASE = 1,
    NEXUS_PATH_TYPE_LISTENER = 2,
    NEXUS_PATH_MAXVAL
} nexus_path_type_t;

    
/*
 * nexus_arch_type_t
 */
typedef enum _nexus_arch_type_t
{
    NEXUS_ARCH_TYPE_UNKNOWN,
    NEXUS_ARCH_TYPE_ANY,
    NEXUS_ARCH_TYPE_SUN3,
    NEXUS_ARCH_TYPE_SUN4,
    NEXUS_ARCH_TYPE_NEXT040,
    NEXUS_ARCH_TYPE_NEXTX86,
    NEXUS_ARCH_TYPE_NEXTPA,
    NEXUS_ARCH_TYPE_NEXTSPARC,
    NEXUS_ARCH_TYPE_RS6000,
    NEXUS_ARCH_TYPE_PARAGON,
    NEXUS_ARCH_TYPE_NTX86,
    NEXUS_ARCH_TYPE_CRAYC90
} nexus_arch_type_t;


/*
 * nexus_class_type_t
 *
 * If you add an entry to this table, you must also add an
 * entry in pr_iface.c:class_type_pointer_sizes[].
 */
typedef int nexus_class_type_t;

#define NEXUS_CLASS_TYPE_32BIT_LE	1
		/* little endian
		 * 16  bit short
		 * 32  bit integer
		 * 32  bit long
		 * 32  bit pointer
		 * 32  bit IEEE float
		 * 64  bit IEEE double
		 */
#define NEXUS_CLASS_TYPE_32BIT_BE	2
		/* big endian
		 * 16  bit short
		 * 32  bit integer
		 * 32  bit long
		 * 32  bit pointer
		 * 32  bit IEEE float
		 * 64  bit IEEE double
		 */
#define NEXUS_CLASS_TYPE_64BIT_LE	3
		/* little endian
		 * 16  bit short
		 * 32  bit integer
		 * 64  bit long
		 * 64  bit pointer
		 * 32  bit IEEE float
		 * 64 bit IEEE double
		 */
#define NEXUS_CLASS_TYPE_64BIT_BE	4
		/* big endian
		 * 16  bit short
		 * 32  bit integer
		 * 64  bit long
		 * 64  bit pointer
		 * 32  bit IEEE float
		 * 64 bit IEEE double
		 */
#define NEXUS_CLASS_TYPE_CRAYC90	5
#define NEXUS_CLASS_TYPE_UNKNOWN32	6
		/* 32 bit pointer */
#define NEXUS_CLASS_TYPE_UNKNOWN64	7
		/* 64 bit pointer */
#define NEXUS_CLASS_TYPE_MAXVAL		8


/*
 * NEXUS_PROTO_TYPE_*
 *
 * Each different Nexus protocol module should have a unique
 * type identifier for all times...
 */
typedef enum _nexus_proto_type_t
{
    NEXUS_PROTO_TYPE_LOCAL	= 0,
    NEXUS_PROTO_TYPE_TCP	= 1,
    NEXUS_PROTO_TYPE_SHMEM	= 2,
    NEXUS_PROTO_TYPE_INX	= 3,
    NEXUS_PROTO_TYPE_PVM	= 4,
    NEXUS_PROTO_TYPE_EUI	= 5,
    NEXUS_PROTO_TYPE_ATM	= 6,
    NEXUS_PROTO_TYPE_MPL	= 7,
    NEXUS_PROTO_TYPE_MPINX	= 8,
    NEXUS_PROTO_TYPE_UDP	= 9,
    NEXUS_PROTO_TYPE_MN_UDP	= 10,
    NEXUS_PROTO_TYPE_MAXVAL, /* do not assign a value.. should float to one plus maximum valid number */
    NEXUS_PROTO_TYPE_INTERNAL1	= 1000001,
    NEXUS_PROTO_TYPE_INTERNAL2	= 1000002
} nexus_proto_type_t;

#ifdef NEXUS_DEBUG
#include "dbgstate.h"
#endif /* NEXUS_DEBUG */

EXTERN_C_END


#ifdef NEXUS_LITE
#include <nexuslite_config.h>
#else
#include <nexus_config.h>
#endif

#if defined(NEXUS_LITE) && !defined(PORTS0_LITE)
#define PORTS0_LITE
#endif

/*
 * Nexus should use macros internally when calling other
 * Nexus routines.  And other code using Nexus should probably
 * use macros wherever possible to avoid function call overhead.
 * But if you do not want Nexus to use macros,
 * and instead to always use real function calls, then
 * define NEXUS_DO_NOT_USE_MACROS.
 * For example, this could be defined in nexus_config.h
 * or in user code before including nexus.h.
 */
#ifdef NEXUS_DO_NOT_USE_MACROS
#undef NEXUS_USE_MACROS
#endif

#ifndef NEXUS_ARCH_TYPE
#define NEXUS_ARCH_TYPE NEXUS_ARCH_TYPE_UNKNOWN
#endif

#ifndef NEXUS_CLASS_TYPE
#define NEXUS_CLASS_TYPE NEXUS_CLASS_TYPE_UNKNOWN
#endif

EXTERN_C_BEGIN

#define	NEXUS_TRUE		1
#define	NEXUS_FALSE		0

#define NEXUS_MAX(V1,V2) (((V1) > (V2)) ? (V1) : (V2))
#define NEXUS_MIN(V1,V2) (((V1) < (V2)) ? (V1) : (V2))

#ifdef NEXUS_DEFINE_GLOBALS
#define NEXUS_GLOBAL
#else
#define NEXUS_GLOBAL extern
#endif

#ifndef NEXUS_DEFAULT_LIBA_SIZE
#define NEXUS_DEFAULT_LIBA_SIZE (2*sizeof(unsigned long))
#endif

#define NEXUS_MAX_HANDLER_NAME_LENGTH 255

#define NEXUS_MAX_EXECUTABLE_PATH_LENGTH 1023

#define NEXUS_MAX_DEBUG_LEVELS 10

#ifdef NEXUS_DEBUG
#define NexusAssert(assertion) \
    if (!(assertion)) \
    { \
        nexus_fatal("Assertion " #assertion " failed in file %s at line %d\n",\
		   __FILE__, __LINE__); \
    }
#define NexusAssert2(assertion, print_args) \
    if (!(assertion)) \
    { \
        nexus_fatal("Assertion " #assertion " failed in file %s at line %d: %s", \
		   __FILE__, __LINE__, nexus_assert_sprintf print_args); \
    }
#else /* NEXUS_DEBUG */
#define NexusAssert(assertion)
#define NexusAssert2(assertion, print_args)
#endif /* NEXUS_DEBUG */


#define NexusMalloc(Func, Var, Type, Size) \
{ \
    if ((Size) > 0) \
    { \
	if (((Var) = (Type) ports0_malloc (Size)) == (Type) NULL) \
	{ \
	    nexus_fatal("%s: malloc of size %d failed for %s %s in file %s line %d\n", \
                        #Func, (Size), #Type, #Var, __FILE__, __LINE__); \
	} \
    } \
    else \
    { \
	(Var) = (Type) NULL; \
    } \
}

#define NexusFree(Ptr) \
{ \
    if ((Ptr) != NULL) \
    { \
	ports0_free(Ptr); \
    } \
}


/*
 * Some forward typedef declarations...
 */
typedef struct _nexus_proto_funcs_t	nexus_proto_funcs_t;
typedef struct _nexus_buffer_funcs_t	nexus_buffer_funcs_t;
typedef struct _nexus_buffer_t *	nexus_buffer_t;
typedef struct _nexus_stashed_buffer_t *nexus_stashed_buffer_t;
typedef struct _nexus_global_pointer_t	nexus_global_pointer_t;
typedef struct _nexus_proto_t		nexus_proto_t;
typedef struct _nexus_mi_proto_t	nexus_mi_proto_t;


/*
 * Handler stuff
 */
typedef void (*nexus_non_threaded_handler_func_t)(void *address,
						  nexus_buffer_t *buffer);
typedef void (*nexus_threaded_handler_func_t)(void *address,
					      nexus_stashed_buffer_t *buffer);
typedef void (*nexus_unknown_handler_func_t)(void *address,
					     nexus_buffer_t *buffer,
					     char *handler_name,
					     int handler_id);
typedef void (*nexus_exit_handler_func_t)(void);
typedef int  (*nexus_acquired_as_node_handler_func_t)(void);
typedef void *nexus_handler_func_t;

typedef enum _nexus_handler_type_t
{
    NEXUS_HANDLER_TYPE_THREADED,
    NEXUS_HANDLER_TYPE_NON_THREADED
} nexus_handler_type_t;

typedef struct _nexus_handler_t {
    char *			name;
    int				id;  
    nexus_handler_type_t	type;
    nexus_handler_func_t	func;	/* nexus_non_threaded_handler_func_t */
					/* or nexus_threaded_handler_func_t  */
} nexus_handler_t;


/*
 * Other typedefs...
 */
typedef void *(*nexus_thread_func_t)(void *user_arg);
typedef int		nexus_bool_t;
typedef unsigned char	nexus_byte_t;


/*
 * The user must provide a variable containing an initial set of
 * modules that are staticly linked into the executable.
 * This variables contains an array of the triples:
 *	{family_name, module_name, info_function_ptr}
 * Where:
 *   family_name is:
 *	"startups"	startup module
 *	"protocols"	protocol module
 *	"database"	database module
 *   module_name is a unique (within the family) string name for the module.
 *   info_function_ptr is a function pointer to the module's info function.
 *
 * The general form of a module information function is
 *	_nx_<family_name>_<module_name>_info().
 * So the user might supply a table that looks like:
 *
 * extern void *_nx_st_rsh_info(void);
 * extern void *_nx_pr_tcp_info(void);
 * nexus_module_list_t nexus_module_list[] =
 * {
 *     {"startups", "rsh", _nx_st_rsh_info},
 *     {"protocols", "tcp", _nx_pr_tcp_info},
 *     {NULL, NULL, NULL},
 * };
 *
 * This would specify that the rsh startup module, and the tcp protocol
 * module are staticly linked into the program.
 *
 * Note: On machines that provide dynamic linking, this list can be
 * empty.  All necessary modules will be dynamically loaded.  However,
 * linking some in using this list may provide for quicker startup.
 * Also, some startup modules need to be staticly linked to bootstrap
 * properly.
 */
typedef struct _nexus_module_list_t
{
    char *family_name;
    char *module_name;
    void *(*info_func)(void);
} nexus_module_list_t;
/*
extern nexus_module_list_t nexus_module_list[];
*/

EXTERN_C_END


/*********************************************************************
 * Include ports0
 */
#if defined(NEXUS_USE_MACROS) && !defined(PORTS0_USE_MACROS)
#define PORTS0_USE_MACROS
#endif
#include "ports0.h"


/*********************************************************************
 * Include sanity checking header stuff
 */
#include "nx_sanity.h"


#ifndef NEXUS_LITE

/*********************************************************************
 * Include the thread module's external header, which should contain:
 *   1) #include's for system thread headers
 *   2) typedefs for external nexus types
 *   3) macros for external nexus routines
 */
#ifndef NEXUS_TH_HEADER
#define NEXUS_TH_HEADER "th.h"
#endif

#include NEXUS_TH_HEADER

#endif /* NEXUS_LITE */


/*********************************************************************
 * Some typdefs
 */

EXTERN_C_BEGIN

#ifdef NEXUS_LITE
/*
 * For NexusLite we need to dummy up some simple versions
 * of the mutex and cond stuff.
 */
typedef int	nexus_mutex_t;
typedef int	nexus_cond_t;
typedef int	nexus_mutexattr_t;
typedef int	nexus_condattr_t;
#endif /* NEXUS_LITE */


/*
 * nexus_create_context_handle_t
 */
typedef struct _nexus_create_context_handle_t
{
    int				type;
    nexus_mutex_t		mutex;
    nexus_cond_t		cond;
    int				total;
    int				checked_in;
    int				next_gp;
    nexus_global_pointer_t **	gp;
    int **			rc;
} nexus_create_context_handle_t;


/*
 * Table containing sizes of the various data types, in bytes.
 */
typedef struct _nexus_sizeof_table_t
{
    int float_size;
    int double_size;
    int short_size;
    int u_short_size;
    int int_size;
    int u_int_size;
    int long_size;
    int u_long_size;
    int char_size;
    int u_char_size;
    int byte_size;
} nexus_sizeof_table_t;


/*
 * nexus_buffer_t
 *
 * This is the only part of a nexus_buffer_t that
 * needs to be exported to the user.  (i.e., There is no need
 * to export specific protocol buffer information to the user.)
 * It is the part of nexus_buffer_t that is common to the buffer_t's
 * used by each different protocol.
 *
 * This part of the nexus_buffer_t needs to live here so that
 * macros below can work.
 *
 * All protocol buffer_t's (i.e., tcp_buffer_t) should start with
 * the same fields as nexus_buffer_t.  They should use
 * typecasts to convert to their own local type.
 *
 * Technically we are using this as a union of all of the different
 * buffer_t's of the protocols.  However, since the set of protocols
 * to use in a particular executable should be selectable at
 * link time, we cannot use an actual union, but must rely on type
 * casting instead.
 */
struct _nexus_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif
    nexus_buffer_funcs_t *	funcs;
};

struct _nexus_stashed_buffer_t
{
#ifdef NEXUS_DEBUG
    int				magic;
#endif
    nexus_buffer_funcs_t *	funcs;
};


/*
 * nexus_proto_t
 *
 * nexus_proto_t is similar to nexus_buffer_t.
 * Each protocol will have to add data to its version of a proto_t.
 * This structure is the common root to all of them -- all proto_t's
 * should start with these same fields.
 *
 * This part of the nexus_proto_t needs to live here so that
 * macros below can work.
 */
struct _nexus_proto_t
{
    nexus_proto_type_t	 type;
    nexus_proto_funcs_t *funcs;
};


/*
 * nexus_mi_proto_t
 *
 * This structure contains a "machine independent" representation
 * of the protocols that can be used by a global pointer (gp) to
 * deliver a message for that gp.
 *
 * The protocols are stored in the packed byte array.
 * The contents of this byte array are:
 *	node number (4 byte, big endian integer)
 *		The node number to which this mi_proto points.
 *	context differentiator (4 byte, big endian integer)
 *		When contexts are implemented as processes, it is
 *		possible for the node number, node name, and
 *		context address (stored in the gp->liba)
 *		to be identical between two gp's
 *		even though they really point to different
 * 		contexts.  This can happen because the context
 *		address which is stored in the gp->liba is _not_
 *		a unique context id within this node, but
 *		rather is only unique within a single process that
 *		is part of a node.
 *		The context differentiator disambiguates this situation.
 *		Generally it contains the process id of the process
 *		representing this mi_proto.
 *	node number (null terminated string)
 *	protocols: 0 or more entries of the following form
 *	    protocol type (2 byte, big endian integer)
 *	    size of next array (2 byte, big endian integer)
 *	    protocol specific byte array
 *		This contains information that a protocol module can
 *		use to deliver a message.
 *
 * The 'proto' field is an instantiated version on one of the protocols
 * in the byte 'array'.
 *
 * This structure is _not_ reference counted, though the nexus_proto_t
 * object stored in this structure is.  The nexus_proto_t reference
 * count is of the number of global pointers which point through this
 * nexus_mi_proto_t structure to the given nexus_proto_t object.
 */
struct _nexus_mi_proto_t
{
    nexus_proto_t *		proto;
    struct _nexus_mi_proto_t *	next;
    int				size;
    nexus_byte_t *		array;
};


/*
 * nexus_global_pointer_t
 *
 * A global pointer points to an address within a particular context.
 * That context obviously lives within some process on some machine.
 * There may or may not be other contexts within that same process.
 *
 * The 'mi_proto' field provides a description of how
 * to get to the correct process for a given context.
 * It also contains the node name and number to which this global
 * pointer points.
 *
 * The 'liba' (locally interpreted byte array) is a byte array which
 * only the process pointed to by the 'proto' needs to intepret.
 * The first two words of this structure (where the word size depends
 * on the class_type) contain the context and address pointed to by
 * this global pointer, respectively.  These fields cannot be used
 * locally (for example, we do not know if they are big or little
 * endian), except that they can be used for byte comparison to test
 * if two global pointers point to the same context and/or address.
 */
struct _nexus_global_pointer_t
{
    nexus_mi_proto_t *	mi_proto;
    nexus_class_type_t  class_type;
    unsigned int	liba_is_inline	: 1;
    int			liba_size	: 16 /*(sizeof(int) - 1)*/;
#ifdef NEXUS_PROFILE
    int			node_id;
    int			context_id;
#endif /* NEXUS_PROFILE */
    union
    {
	nexus_byte_t *	pointer;			/* liba_is_inline==0 */
	nexus_byte_t	array[NEXUS_DEFAULT_LIBA_SIZE];	/* liba_is_inline==1 */
    } liba;
};


/*
 * nexus_node_t
 *
 * An array of these is returned by nexus_init() on the master node,
 * and by nexus_acquire_nodes().
 */
typedef struct _nexus_node_t
{
    char *			name;
    int				number;
    nexus_global_pointer_t	gp;
    int				return_code;
} nexus_node_t;

/* Values for nexus_node_t.return_code */
#define NEXUS_NODE_NEW		0
#define NEXUS_NODE_OLD_SAME_EXE	1
#define NEXUS_NODE_OLD_DIFF_EXE	2


#if defined(NEXUS_THREAD_SAFE_STDIO) || defined(NEXUS_LITE)
#define nexus_stdio_lock()
#define nexus_stdio_unlock()
#else
NEXUS_GLOBAL nexus_mutex_t nexus_stdio_mutex;
#define nexus_stdio_lock() nexus_mutex_lock(&nexus_stdio_mutex);
#define nexus_stdio_unlock() nexus_mutex_unlock(&nexus_stdio_mutex);
#endif /* NEXUS_THREAD_SAFE_STDIO */

/*********************************************************************
 * 		Function indirection table declarations
 *********************************************************************/

/*
 * This is a function indirection table used in conn_t's
 */
struct _nexus_proto_funcs_t
{
    nexus_proto_type_t	(*proto_type)(void);
    void		(*init)(int *argc, char ***argv);
    void		(*shutdown)(nexus_bool_t shutdown_others);
    void		(*abort)(void);
    void		(*poll)(void);
    void		(*blocking_poll)(void);
    void		(*init_remote_service_request)(nexus_buffer_t *buffer,
						   nexus_global_pointer_t *gp,
						       char *handler_name,
						       int handler_id);
    void                (*increment_reference_count)(nexus_proto_t *proto);
    void		(*decrement_reference_count)(nexus_proto_t *proto);
    void		(*get_my_mi_proto)(nexus_byte_t **array,
					   int *size);
    nexus_bool_t	(*construct_from_mi_proto)(nexus_proto_t **proto,
						   nexus_mi_proto_t *mi_proto,
						   nexus_byte_t *array,
						   int size);
    int			(*test_proto)(nexus_proto_t *proto);
};


/*
 * This is a function indirection table used in nexus_buffer_t's
 */
struct _nexus_buffer_funcs_t
{
    void		(*set_buffer_size)(nexus_buffer_t *buffer,
					   int size, int n_elements);
    int			(*check_buffer_size)(nexus_buffer_t *buffer,
					     int slack, int increment);
    int			(*send_remote_service_request)
						(nexus_buffer_t *buffer);
    int			(*send_urgent_remote_service_request)
						(nexus_buffer_t *buffer);
    void		(*free_buffer)(nexus_buffer_t *buffer);
    void		(*stash_buffer)(nexus_buffer_t *buffer,
				       nexus_stashed_buffer_t *stashed_buffer);
    void		(*free_stashed_buffer)
				      (nexus_stashed_buffer_t *stashed_buffer);

    int (*sizeof_float)(nexus_buffer_t *buffer, int count);
    int (*sizeof_double)(nexus_buffer_t *buffer, int count);
    int (*sizeof_short)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_short)(nexus_buffer_t *buffer, int count);
    int (*sizeof_int)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_int)(nexus_buffer_t *buffer, int count);
    int (*sizeof_long)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_long)(nexus_buffer_t *buffer, int count);
    int (*sizeof_char)(nexus_buffer_t *buffer, int count);
    int (*sizeof_u_char)(nexus_buffer_t *buffer, int count);
    int (*sizeof_byte)(nexus_buffer_t *buffer, int count);

    void (*put_float)(nexus_buffer_t *buffer,
		      float *data, int count);
    void (*put_double)(nexus_buffer_t *buffer,
		       double *data, int count);
    void (*put_short)(nexus_buffer_t *buffer,
		      short *data, int count);
    void (*put_u_short)(nexus_buffer_t *buffer,
			unsigned short *data, int count);
    void (*put_int)(nexus_buffer_t *buffer,
		    int *data, int count);
    void (*put_u_int)(nexus_buffer_t *buffer,
		      unsigned int *data, int count);
    void (*put_long)(nexus_buffer_t *buffer,
		     long *data, int count);
    void (*put_u_long)(nexus_buffer_t *buffer,
		       unsigned long *data, int count);
    void (*put_char)(nexus_buffer_t *buffer,
		     char *data, int count);
    void (*put_u_char)(nexus_buffer_t *buffer,
		       unsigned char *data, int count);
    void (*put_byte)(nexus_buffer_t *buffer,
		     unsigned char *data, int count);

    void (*get_float)(nexus_buffer_t *buffer,
		      float *dest, int count);
    void (*get_double)(nexus_buffer_t *buffer,
		       double *dest, int count);
    void (*get_short)(nexus_buffer_t *buffer,
		      short *dest, int count);
    void (*get_u_short)(nexus_buffer_t *buffer,
			unsigned short *dest, int count);
    void (*get_int)(nexus_buffer_t *buffer,
		    int *dest, int count);
    void (*get_u_int)(nexus_buffer_t *buffer,
		      unsigned int *dest, int count);
    void (*get_long)(nexus_buffer_t *buffer,
		     long *dest, int count);
    void (*get_u_long)(nexus_buffer_t *buffer,
		       unsigned long *dest, int count);
    void (*get_char)(nexus_buffer_t *buffer,
		     char *dest, int count);
    void (*get_u_char)(nexus_buffer_t *buffer,
		       unsigned char *dest, int count);
    void (*get_byte)(nexus_buffer_t *buffer,
		     unsigned char *dest, int count);

    void (*get_stashed_float)(nexus_stashed_buffer_t *buffer,
			      float *dest, int count);
    void (*get_stashed_double)(nexus_stashed_buffer_t *buffer,
			       double *dest, int count);
    void (*get_stashed_short)(nexus_stashed_buffer_t *buffer,
			      short *dest, int count);
    void (*get_stashed_u_short)(nexus_stashed_buffer_t *buffer,
				unsigned short *dest, int count);
    void (*get_stashed_int)(nexus_stashed_buffer_t *buffer,
			    int *dest, int count);
    void (*get_stashed_u_int)(nexus_stashed_buffer_t *buffer,
			      unsigned int *dest, int count);
    void (*get_stashed_long)(nexus_stashed_buffer_t *buffer,
			     long *dest, int count);
    void (*get_stashed_u_long)(nexus_stashed_buffer_t *buffer,
			       unsigned long *dest, int count);
    void (*get_stashed_char)(nexus_stashed_buffer_t *buffer,
			     char *dest, int count);
    void (*get_stashed_u_char)(nexus_stashed_buffer_t *buffer,
			       unsigned char *dest, int count);
    void (*get_stashed_byte)(nexus_stashed_buffer_t *buffer,
			     unsigned char *dest, int count);
};



/*********************************************************************
 * Declarations for functions that must be supplied by the package
 *********************************************************************/

extern int	NexusBoot(void);


/*********************************************************************
 * Declarations for functions that may optionally be supplied by
 * the package by registering them as handlers.
 *********************************************************************/

extern void	NexusExit(void);
extern void	NexusUnknownHandler(void *address,
				    nexus_buffer_t *buffer,
				    char *handler_name,
				    int handler_id);
extern int	NexusAcquiredAsNode(void);

#define NEXUS_EXIT_HASH 941
#define NEXUS_UNKNOWN_HANDLER_HASH 964
#define NEXUS_ACQUIRED_AS_NODE_HASH 894
#define NEXUS_LISTENER_PATH_HASH 745


/*********************************************************************
 * 		Exported interface declarations
 *********************************************************************/
extern int	nexus_user_iargc(void);
extern char *	nexus_user_getarg(int i);
extern int	nexus_package_iargc(void);
extern char *	nexus_package_getarg(int i);
extern void	nexus_get_argc_and_argv(int *argc, char ***argv);
extern void     nexus_set_path(nexus_path_type_t, char *);

extern void	nexus_init(int *argc,
			   char ***argv,
			   char *args_env_variable,
			   char *package_designator,
			   int (*package_args_init_func)(int *argc,
							 char ***argv),
			   void (*usage_message_func)(void),
			   int (*new_process_params_func)(char *buf, int size),
			   nexus_module_list_t module_list[],
			   nexus_node_t **nodes,
			   int *n_nodes);
extern void	nexus_start(void);
extern void	nexus_start_nonblocking(void);
extern void	nexus_shutdown_nonexiting(void);
extern void	nexus_exit(int rc, int shutdown);
extern void	nexus_shutdown(void);
extern void	nexus_abort(void);

extern void	nexus_destroy_current_context(
				int called_from_non_threaded_handler);
extern void	nexus_acquire_nodes(char *node_name,
				    int node_number,
				    int count,
				    char *directory_path,
				    char *executable_path,
				    nexus_node_t **nodes,
				    int *n_nodes);

extern void	nexus_init_create_context_handle(
				nexus_create_context_handle_t *contexts,
                                int n_contexts);
extern void	nexus_create_context(nexus_global_pointer_t *node_gp,
				     char *executable_path,
				     int size,
				     nexus_global_pointer_t *new_context_gp,
				     int *return_code,
				     nexus_create_context_handle_t *contexts);
extern void	nexus_create_context_wait(
				nexus_create_context_handle_t *contexts);
extern void *	nexus_context_initial_segment(void);
extern void *	nexus_malloc(size_t size);
extern void	nexus_free(void *data_segment);

extern void	nexus_poll(void);
extern void	nexus_poll_profile(int id);
extern void	nexus_blocking_poll(void);
extern void	nexus_global_pointer(nexus_global_pointer_t *new_gp,
				     void *address);
extern void *	nexus_convert_global_pointer_address(
						nexus_global_pointer_t *gp);
extern void	nexus_destroy_global_pointer(nexus_global_pointer_t *gp);
extern void     nexus_copy_global_pointer(nexus_global_pointer_t *gp_out,
					  nexus_global_pointer_t *gp_in);
extern int	nexus_sizeof_global_pointer(nexus_buffer_t *buffer,
					    nexus_global_pointer_t *gp,
					    int count,
					    int *n_elements);
extern void	nexus_put_global_pointer(nexus_buffer_t *buffer,
					 nexus_global_pointer_t *gp,
					 int count);
extern void	nexus_get_global_pointer(nexus_buffer_t *buffer,
					 nexus_global_pointer_t *gp,
					 int count);
extern void	nexus_get_stashed_global_pointer(
						nexus_stashed_buffer_t *buffer,
						nexus_global_pointer_t *gp,
						int count);


extern int	nexus_handler_hash(char *name);
extern void	nexus_register_handlers(nexus_handler_t *handlers);
extern void	nexus_substitute_handler(char *name,
					 int id,
					 nexus_handler_type_t new_type,
					 void *new_func,
					 nexus_handler_type_t *old_type,
					 void **old_func);

extern nexus_arch_type_t	nexus_node_type(nexus_global_pointer_t *gp);
extern nexus_class_type_t	nexus_node_class(nexus_global_pointer_t *gp);
extern nexus_global_pointer_t *	nexus_node_exists(char *hostname,
						  char *path);
extern int nexus_same_context(nexus_global_pointer_t *gp1,
			      nexus_global_pointer_t *gp2);
extern int nexus_same_global_pointer(nexus_global_pointer_t *gp1,
				     nexus_global_pointer_t *gp2);
extern int nexus_global_pointer_to_current_context(nexus_global_pointer_t *gp);

extern int nexus_master_node(void);


/*
 * fault tolerance stuff
 */
extern void nexus_enable_fault_tolerance(int (*callback_func)(void *user_arg,
							      int fault_code),
					 void *callback_func_user_arg);
extern int	nexus_test_global_pointer(nexus_global_pointer_t *gp);

#define NEXUS_FAULT_NONE			0
#define NEXUS_FAULT_PROCESS_DIED		1
#define NEXUS_FAULT_PROCESS_SHUTDOWN_ABNORMALLY	2
#define NEXUS_FAULT_PROCESS_SHUTDOWN_NORMALLY	3
#define NEXUS_FAULT_ATTACHER_DIED		4
#define NEXUS_FAULT_CONNECT_FAILED		5
#define NEXUS_FAULT_BAD_PROTOCOL		6

extern char *nexus_fault_strings[];


/*
 * args.c
 */
extern int	nexus_find_argument(int *argc,
				    char ***argv,
				    char *arg,
				    int count);
extern void	nexus_remove_arguments(int *argc,
				       char ***argv,
				       int arg_num,
				       int count);
extern void	nexus_get_retained_arguments(int **argc, char ****argv);

/*
 * db_iface.c
 */
extern char *   nexus_database_lookup(char *node_name,
				      int node_number,
				      char *key);
extern void     nexus_database_free(char *value);

/*
 * error.c
 */
extern void		nexus_silent_fatal(void);
#if defined(__STDC__) || defined(__cplusplus)
extern void		nexus_fatal(char *msg, ...);
extern void		nexus_error(char *msg, ...);
extern void		nexus_warning(char *msg, ...);
extern void		nexus_notice(char *msg, ...);
extern void		nexus_printf(char *msg, ...);
extern void		nexus_perror(char *msg, ...);
extern void		nexus_fatal_perror(char *msg, ...);
extern char *		nexus_assert_sprintf(char *msg, ...);
#else /* __STDC__ */
/*
 * These have variable number of parameters, so must be defined this way
 * to get the compiler to work correctly.
 */
extern void		nexus_fatal();
extern void		nexus_error();
extern void		nexus_warning();
extern void		nexus_notice();
extern void		nexus_printf();
extern void		nexus_perror();
extern void		nexus_fatal_perror();
extern char *		nexus_assert_sprintf();
#endif /* __STDC__ */


/*
 * md_*.c
 */
extern void		nexus_kill(int pid);
extern void 		nexus_usleep(long usecs);
extern double		nexus_wallclock(void);
extern void		nexus_pause(void);


/*
 * pr_tcp.c
 */
extern int		nexus_allow_attach(unsigned short *port,
					   char **host,
					   int (*approval_func)(void *user_arg,
								char *url,
						   nexus_global_pointer_t *gp),
					   void *approval_func_user_arg
					   );
extern void		nexus_disallow_attach(unsigned short port);
extern int		nexus_attach(char *url,
				     nexus_global_pointer_t *gp);
extern int		nexus_split_nexus_url(char *url,
					      char **host,
					      unsigned short *port,
					      char ***specifiers);
extern void		nexus_split_nexus_url_free(char **host,
						   char ***specifiers);

/*
 * nexus_register_fd_notifiers() registers functions that
 * Nexus should call to register any fd's that it uses.
 *
 * Note: This is _not_ a general solution.  It exists so that nperl
 * does not have to use a patched version of nexus.  It _will_ change
 * sometime in the future.  You should not rely on this function.
 */
typedef void (*nexus_fd_notifier_func_t)(int fd);
extern void	nexus_register_fd_notifiers(nexus_fd_notifier_func_t add,
					    nexus_fd_notifier_func_t remove);



/*
 * th_*.c
 */
extern nexus_bool_t		nexus_preemptive_threads(void);


/*
 * utils.c
 */
extern void		nexus_ids(int *node_id,
				  int *context_id,
				  int *thread_id);



/*********************************************************************
 * 		Exported interface routines that can be macros
 *********************************************************************/

#define nexus_macro_set_package_id(package_id_base) \
    ports0_set_package_id(package_id_base)
#define nexus_macro_get_package_id() \
    ports0_get_package_id()
#define	nexus_macro_null_global_pointer(gp) \
{ \
    (gp)->mi_proto = (nexus_mi_proto_t *) NULL; \
}
#define nexus_macro_is_null_global_pointer(gp) \
    ((gp)->mi_proto == (nexus_mi_proto_t *) NULL)
#define nexus_macro_init_remote_service_request(buf, gp, h_name, h_id) \
    ((gp)->mi_proto->proto->funcs->init_remote_service_request)(buf, gp, \
								h_name, h_id)
#ifndef NEXUS_USE_MACROS
extern int	nexus_set_package_id(char *package_id_base);
extern char *	nexus_get_package_id(void);
extern void	nexus_null_global_pointer(nexus_global_pointer_t *gp);
extern nexus_bool_t	nexus_is_null_global_pointer(nexus_global_pointer_t *gp);
extern void nexus_init_remote_service_request(nexus_buffer_t *buffer,
					      nexus_global_pointer_t *gp,
					      char *handler_name,
					      int handler_id);
#else  /* NEXUS_USE_MACROS */
#define nexus_set_package_id(package_id_base) \
    nexus_macro_set_package_id(package_id_base)
#define nexus_get_package_id() \
    nexus_macro_get_package_id()
#define	nexus_null_global_pointer(gp) \
    nexus_macro_null_global_pointer(gp)
#define nexus_is_null_global_pointer(gp) \
    nexus_macro_is_null_global_pointer(gp)
#define nexus_init_remote_service_request(buf, gp, h_name, h_id) \
    nexus_macro_init_remote_service_request(buf, gp, h_name, h_id)
#endif /* NEXUS_USE_MACROS */



#define nexus_macro_set_buffer_size(buffer, size, n_elements) \
    ((*buffer)->funcs->set_buffer_size)(buffer, size, n_elements)
#define nexus_macro_check_buffer_size(buffer, slack, increment) \
    ((*buffer)->funcs->check_buffer_size)(buffer, slack, increment)
#define nexus_macro_send_remote_service_request(buffer) \
    ((*buffer)->funcs->send_remote_service_request)(buffer)
#define nexus_macro_send_urgent_remote_service_request(buffer) \
    ((*buffer)->funcs->send_urgent_remote_service_request)(buffer)
#define nexus_macro_free_buffer(buffer) \
{ \
    if (*buffer) /* *buffer==NULL if the buffer has been stashed */ \
    { \
	((*buffer)->funcs->free_buffer)(buffer); \
    } \
}
#define nexus_macro_stash_buffer(buffer, stashed_buffer) \
    ((*buffer)->funcs->stash_buffer)(buffer, stashed_buffer)
#define nexus_macro_free_stashed_buffer(stashed_buffer) \
    ((*stashed_buffer)->funcs->free_stashed_buffer)(stashed_buffer)
#ifndef NEXUS_USE_MACROS
extern void		nexus_set_buffer_size(nexus_buffer_t *buffer,
					      int size,
					      int n_elements);
extern int		nexus_check_buffer_size(nexus_buffer_t *buffer,
						int slack,
						int increment);
extern int		nexus_send_remote_service_request(
						nexus_buffer_t *buffer);
extern int		nexus_send_urgent_remote_service_request(
						nexus_buffer_t *buffer);
extern void		nexus_free_buffer(nexus_buffer_t *buffer);
extern void		nexus_stash_buffer(nexus_buffer_t *buffer,
				   nexus_stashed_buffer_t *stashed_buffer);
extern void		nexus_free_stashed_buffer(
				   nexus_stashed_buffer_t *stashed_buffer);
#else  /* NEXUS_USE_MACROS */
#define nexus_set_buffer_size(buffer, size, n_elements) \
    nexus_macro_set_buffer_size(buffer, size, n_elements)
#define nexus_check_buffer_size(buffer, slack, increment) \
    nexus_macro_check_buffer_size(buffer, slack, increment)
#define nexus_send_remote_service_request(buffer) \
    nexus_macro_send_remote_service_request(buffer)
#define nexus_send_urgent_remote_service_request(buffer) \
    nexus_macro_send_urgent_remote_service_request(buffer)
#define nexus_free_buffer(buffer) \
    nexus_macro_free_buffer(buffer)
#define nexus_stash_buffer(buffer, stashed_buffer) \
    nexus_macro_stash_buffer(buffer, stashed_buffer)
#define nexus_free_stashed_buffer(stashed_buffer) \
    nexus_macro_free_stashed_buffer(stashed_buffer)
#endif /* NEXUS_USE_MACROS */



#define nexus_macro_sizeof_float(buffer, count) \
    ((*buffer)->funcs->sizeof_float)(buffer, count)
#define nexus_macro_sizeof_double(buffer, count) \
    ((*buffer)->funcs->sizeof_double)(buffer, count)
#define nexus_macro_sizeof_short(buffer, count) \
    ((*buffer)->funcs->sizeof_short)(buffer, count)
#define nexus_macro_sizeof_u_short(buffer, count) \
    ((*buffer)->funcs->sizeof_u_short)(buffer, count)
#define nexus_macro_sizeof_int(buffer, count) \
    ((*buffer)->funcs->sizeof_int)(buffer, count)
#define nexus_macro_sizeof_u_int(buffer, count) \
    ((*buffer)->funcs->sizeof_u_int)(buffer, count)
#define nexus_macro_sizeof_long(buffer, count) \
    ((*buffer)->funcs->sizeof_long)(buffer, count)
#define nexus_macro_sizeof_u_long(buffer, count) \
    ((*buffer)->funcs->sizeof_u_long)(buffer, count)
#define nexus_macro_sizeof_char(buffer, count) \
    ((*buffer)->funcs->sizeof_char)(buffer, count)
#define nexus_macro_sizeof_u_char(buffer, count) \
    ((*buffer)->funcs->sizeof_u_char)(buffer, count)
#define nexus_macro_sizeof_byte(buffer, count) \
    ((*buffer)->funcs->sizeof_byte)(buffer, count)
#ifndef NEXUS_USE_MACROS
extern int	nexus_sizeof_float(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_double(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_short(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_short(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_int(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_int(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_long(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_long(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_char(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_u_char(nexus_buffer_t *buffer, int count);
extern int	nexus_sizeof_byte(nexus_buffer_t *buffer, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_sizeof_float(buffer, count) \
    nexus_macro_sizeof_float(buffer, count)
#define nexus_sizeof_double(buffer, count) \
    nexus_macro_sizeof_double(buffer, count)
#define nexus_sizeof_short(buffer, count) \
    nexus_macro_sizeof_short(buffer, count)
#define nexus_sizeof_u_short(buffer, count) \
    nexus_macro_sizeof_u_short(buffer, count)
#define nexus_sizeof_int(buffer, count) \
    nexus_macro_sizeof_int(buffer, count)
#define nexus_sizeof_u_int(buffer, count) \
    nexus_macro_sizeof_u_int(buffer, count)
#define nexus_sizeof_long(buffer, count) \
    nexus_macro_sizeof_long(buffer, count)
#define nexus_sizeof_u_long(buffer, count) \
    nexus_macro_sizeof_u_long(buffer, count)
#define nexus_sizeof_char(buffer, count) \
    nexus_macro_sizeof_char(buffer, count)
#define nexus_sizeof_u_char(buffer, count) \
    nexus_macro_sizeof_u_char(buffer, count)
#define nexus_sizeof_byte(buffer, count) \
    nexus_macro_sizeof_byte(buffer, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_put_float(buffer, data, count) \
    ((*buffer)->funcs->put_float)(buffer, data, count)
#define nexus_macro_put_double(buffer, data, count) \
    ((*buffer)->funcs->put_double)(buffer, data, count)
#define nexus_macro_put_short(buffer, data, count) \
    ((*buffer)->funcs->put_short)(buffer, data, count)
#define nexus_macro_put_u_short(buffer, data, count) \
    ((*buffer)->funcs->put_u_short)(buffer, data, count)
#define nexus_macro_put_int(buffer, data, count) \
    ((*buffer)->funcs->put_int)(buffer, data, count)
#define nexus_macro_put_u_int(buffer, data, count) \
    ((*buffer)->funcs->put_u_int)(buffer, data, count)
#define nexus_macro_put_long(buffer, data, count) \
    ((*buffer)->funcs->put_long)(buffer, data, count)
#define nexus_macro_put_u_long(buffer, data, count) \
    ((*buffer)->funcs->put_u_long)(buffer, data, count)
#define nexus_macro_put_char(buffer, data, count) \
    ((*buffer)->funcs->put_char)(buffer, data, count)
#define nexus_macro_put_u_char(buffer, data, count) \
    ((*buffer)->funcs->put_u_char)(buffer, data, count)
#define nexus_macro_put_byte(buffer, data, count) \
    ((*buffer)->funcs->put_byte)(buffer, data, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_put_float(nexus_buffer_t *buffer,
			    float *data, int count);
extern void nexus_put_double(nexus_buffer_t *buffer,
			     double *data, int count);
extern void nexus_put_short(nexus_buffer_t *buffer,
			    short *data, int count);
extern void nexus_put_u_short(nexus_buffer_t *buffer,
			      unsigned short *data, int count);
extern void nexus_put_int(nexus_buffer_t *buffer,
			  int *data, int count);
extern void nexus_put_u_int(nexus_buffer_t *buffer,
			    unsigned int *data, int count);
extern void nexus_put_long(nexus_buffer_t *buffer,
			   long *data, int count);
extern void nexus_put_u_long(nexus_buffer_t *buffer,
			     unsigned long *data, int count);
extern void nexus_put_char(nexus_buffer_t *buffer,
			   char *data, int count);
extern void nexus_put_u_char(nexus_buffer_t *buffer,
			     unsigned char *data, int count);
extern void nexus_put_byte(nexus_buffer_t *buffer,
			   unsigned char *data, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_put_float(buffer, data, count) \
    nexus_macro_put_float(buffer, data, count)
#define nexus_put_double(buffer, data, count) \
    nexus_macro_put_double(buffer, data, count)
#define nexus_put_short(buffer, data, count) \
    nexus_macro_put_short(buffer, data, count)
#define nexus_put_u_short(buffer, data, count) \
    nexus_macro_put_u_short(buffer, data, count)
#define nexus_put_int(buffer, data, count) \
    nexus_macro_put_int(buffer, data, count)
#define nexus_put_u_int(buffer, data, count) \
    nexus_macro_put_u_int(buffer, data, count)
#define nexus_put_long(buffer, data, count) \
    nexus_macro_put_long(buffer, data, count)
#define nexus_put_u_long(buffer, data, count) \
    nexus_macro_put_u_long(buffer, data, count)
#define nexus_put_char(buffer, data, count) \
    nexus_macro_put_char(buffer, data, count)
#define nexus_put_u_char(buffer, data, count) \
    nexus_macro_put_u_char(buffer, data, count)
#define nexus_put_byte(buffer, data, count) \
    nexus_macro_put_byte(buffer, data, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_get_float(buffer, dest, count) \
    ((*buffer)->funcs->get_float)(buffer, dest, count)
#define nexus_macro_get_double(buffer, dest, count) \
    ((*buffer)->funcs->get_double)(buffer, dest, count)
#define nexus_macro_get_short(buffer, dest, count) \
    ((*buffer)->funcs->get_short)(buffer, dest, count)
#define nexus_macro_get_u_short(buffer, dest, count) \
    ((*buffer)->funcs->get_u_short)(buffer, dest, count)
#define nexus_macro_get_int(buffer, dest, count) \
    ((*buffer)->funcs->get_int)(buffer, dest, count)
#define nexus_macro_get_u_int(buffer, dest, count) \
    ((*buffer)->funcs->get_u_int)(buffer, dest, count)
#define nexus_macro_get_long(buffer, dest, count) \
    ((*buffer)->funcs->get_long)(buffer, dest, count)
#define nexus_macro_get_u_long(buffer, dest, count) \
    ((*buffer)->funcs->get_u_long)(buffer, dest, count)
#define nexus_macro_get_char(buffer, dest, count) \
    ((*buffer)->funcs->get_char)(buffer, dest, count)
#define nexus_macro_get_u_char(buffer, dest, count) \
    ((*buffer)->funcs->get_u_char)(buffer, dest, count)
#define nexus_macro_get_byte(buffer, dest, count) \
    ((*buffer)->funcs->get_byte)(buffer, dest, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_get_float(nexus_buffer_t *buffer,
			    float *dest, int count);
extern void nexus_get_double(nexus_buffer_t *buffer,
			     double *dest, int count);
extern void nexus_get_short(nexus_buffer_t *buffer,
			    short *dest, int count);
extern void nexus_get_u_short(nexus_buffer_t *buffer,
			      unsigned short *dest, int count);
extern void nexus_get_int(nexus_buffer_t *buffer,
			  int *dest, int count);
extern void nexus_get_u_int(nexus_buffer_t *buffer,
			    unsigned int *dest, int count);
extern void nexus_get_long(nexus_buffer_t *buffer,
			   long *dest, int count);
extern void nexus_get_u_long(nexus_buffer_t *buffer,
			     unsigned long *dest, int count);
extern void nexus_get_char(nexus_buffer_t *buffer,
			   char *dest, int count);
extern void nexus_get_u_char(nexus_buffer_t *buffer,
			     unsigned char *dest, int count);
extern void nexus_get_byte(nexus_buffer_t *buffer,
			   unsigned char *dest, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_get_float(buffer, dest, count) \
    nexus_macro_get_float(buffer, dest, count)
#define nexus_get_double(buffer, dest, count) \
    nexus_macro_get_double(buffer, dest, count)
#define nexus_get_short(buffer, dest, count) \
    nexus_macro_get_short(buffer, dest, count)
#define nexus_get_u_short(buffer, dest, count) \
    nexus_macro_get_u_short(buffer, dest, count)
#define nexus_get_int(buffer, dest, count) \
    nexus_macro_get_int(buffer, dest, count)
#define nexus_get_u_int(buffer, dest, count) \
    nexus_macro_get_u_int(buffer, dest, count)
#define nexus_get_long(buffer, dest, count) \
    nexus_macro_get_long(buffer, dest, count)
#define nexus_get_u_long(buffer, dest, count) \
    nexus_macro_get_u_long(buffer, dest, count)
#define nexus_get_char(buffer, dest, count) \
    nexus_macro_get_char(buffer, dest, count)
#define nexus_get_u_char(buffer, dest, count) \
    nexus_macro_get_u_char(buffer, dest, count)
#define nexus_get_byte(buffer, dest, count) \
    nexus_macro_get_byte(buffer, dest, count)
#endif /* NEXUS_USE_MACROS */


#define nexus_macro_get_stashed_float(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_float)(buffer, dest, count)
#define nexus_macro_get_stashed_double(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_double)(buffer, dest, count)
#define nexus_macro_get_stashed_short(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_short)(buffer, dest, count)
#define nexus_macro_get_stashed_u_short(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_short)(buffer, dest, count)
#define nexus_macro_get_stashed_int(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_int)(buffer, dest, count)
#define nexus_macro_get_stashed_u_int(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_int)(buffer, dest, count)
#define nexus_macro_get_stashed_long(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_long)(buffer, dest, count)
#define nexus_macro_get_stashed_u_long(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_long)(buffer, dest, count)
#define nexus_macro_get_stashed_char(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_char)(buffer, dest, count)
#define nexus_macro_get_stashed_u_char(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_u_char)(buffer, dest, count)
#define nexus_macro_get_stashed_byte(buffer, dest, count) \
    ((*buffer)->funcs->get_stashed_byte)(buffer, dest, count)
#ifndef NEXUS_USE_MACROS
extern void nexus_get_stashed_float(nexus_stashed_buffer_t *buffer,
				    float *dest, int count);
extern void nexus_get_stashed_double(nexus_stashed_buffer_t *buffer,
				     double *dest, int count);
extern void nexus_get_stashed_short(nexus_stashed_buffer_t *buffer,
				    short *dest, int count);
extern void nexus_get_stashed_u_short(nexus_stashed_buffer_t *buffer,
				      unsigned short *dest, int count);
extern void nexus_get_stashed_int(nexus_stashed_buffer_t *buffer,
				  int *dest, int count);
extern void nexus_get_stashed_u_int(nexus_stashed_buffer_t *buffer,
				    unsigned int *dest, int count);
extern void nexus_get_stashed_long(nexus_stashed_buffer_t *buffer,
				   long *dest, int count);
extern void nexus_get_stashed_u_long(nexus_stashed_buffer_t *buffer,
				     unsigned long *dest, int count);
extern void nexus_get_stashed_char(nexus_stashed_buffer_t *buffer,
				   char *dest, int count);
extern void nexus_get_stashed_u_char(nexus_stashed_buffer_t *buffer,
				     unsigned char *dest, int count);
extern void nexus_get_stashed_byte(nexus_stashed_buffer_t *buffer,
				   unsigned char *dest, int count);
#else  /* NEXUS_USE_MACROS */
#define nexus_get_stashed_float(buffer, dest, count) \
    nexus_macro_get_stashed_float(buffer, dest, count)
#define nexus_get_stashed_double(buffer, dest, count) \
    nexus_macro_get_stashed_double(buffer, dest, count)
#define nexus_get_stashed_short(buffer, dest, count) \
    nexus_macro_get_stashed_short(buffer, dest, count)
#define nexus_get_stashed_u_short(buffer, dest, count) \
    nexus_macro_get_stashed_u_short(buffer, dest, count)
#define nexus_get_stashed_int(buffer, dest, count) \
    nexus_macro_get_stashed_int(buffer, dest, count)
#define nexus_get_stashed_u_int(buffer, dest, count) \
    nexus_macro_get_stashed_u_int(buffer, dest, count)
#define nexus_get_stashed_long(buffer, dest, count) \
    nexus_macro_get_stashed_long(buffer, dest, count)
#define nexus_get_stashed_u_long(buffer, dest, count) \
    nexus_macro_get_stashed_u_long(buffer, dest, count)
#define nexus_get_stashed_char(buffer, dest, count) \
    nexus_macro_get_stashed_char(buffer, dest, count)
#define nexus_get_stashed_u_char(buffer, dest, count) \
    nexus_macro_get_stashed_u_char(buffer, dest, count)
#define nexus_get_stashed_byte(buffer, dest, count) \
    nexus_macro_get_stashed_byte(buffer, dest, count)
#endif /* NEXUS_USE_MACROS */


#ifdef NEXUS_LITE
/*
 * For NexusLite we need to dummy up some simple versions
 * of the mutex and cond stuff.
 */

#define nexus_macro_mutex_init(M,A)	(*(M) = 0)
#define nexus_macro_mutex_destroy(M)	(*(M) = 0)
#define nexus_macro_mutex_lock(M) \
    ( (*(M)) \
     ? (nexus_fatal("nexus_mutex_lock(): Deadlock detected in file %s at line %d. The mutex at address 0x%lu is already locked.\n", __FILE__, __LINE__, (unsigned long) (M)), 1) \
     : ( ((*(M)) = 1), 0 ) )
#define nexus_macro_mutex_unlock(M)	( ((*(M)) = 0), 0 )
#define nexus_macro_mutex_trylock(M)    nexus_macro_mutex_lock(M)
#define nexus_macro_cond_init(C,A)	(*(C) = 0)
#define nexus_macro_cond_destroy(C)	(*(C) = 0)
#define nexus_macro_cond_wait(C,M) \
    ( ((*(M)) = 0), nexus_blocking_poll(), ((*(M)) = 1), 0 )
#define nexus_macro_cond_signal(C)	(*(C) = 0)
#define nexus_macro_cond_broadcast(C)	(*(C) = 0)

#ifndef NEXUS_USE_MACROS
extern int		nexus_mutex_init(nexus_mutex_t *mutex,
					 nexus_mutexattr_t *attr);
extern int		nexus_mutex_destroy(nexus_mutex_t *mutex);
extern int		nexus_mutex_lock(nexus_mutex_t *mutex);
extern int		nexus_mutex_unlock(nexus_mutex_t *mutex);
extern int              nexus_mutex_trylock(nexus_mutex_t *mutex);
extern int		nexus_cond_init(nexus_cond_t *cond,
					nexus_condattr_t *attr);
extern int		nexus_cond_destroy(nexus_cond_t *cond);
extern int		nexus_cond_wait(nexus_cond_t *cond,
					nexus_mutex_t *mutex);
extern int		nexus_cond_signal(nexus_cond_t *cond);
extern int		nexus_cond_broadcast(nexus_cond_t *cond);
#else  /* NEXUS_USE_MACROS */
#define nexus_mutex_init(M,A) nexus_macro_mutex_init(M,A)
#define nexus_mutex_destroy(M) nexus_macro_mutex_destroy(M)
#define nexus_mutex_lock(M) nexus_macro_mutex_lock(M)
#define nexus_mutex_unlock(M) nexus_macro_mutex_unlock(M)
#define nexus_mutex_trylock(M) nexus_macro_mutex_trylock(M)
#define nexus_cond_init(C,A) nexus_macro_cond_init(C,A)
#define nexus_cond_destroy(C) nexus_macro_cond_destroy(C)
#define nexus_cond_wait(C,M) nexus_macro_cond_wait(C,M)
#define nexus_cond_signal(C) nexus_macro_cond_signal(C)
#define nexus_cond_broadcast(C) nexus_macro_cond_broadcast(C)
#endif /* NEXUS_USE_MACROS */

#endif /* NEXUS_LITE */


EXTERN_C_END

/* 
 * At this point, we have defined all of the macros that can effect
 * the size of structures in nexus.h
 */
#include "nx_ver.h"

#define NEXUS_NO_NODE_NUM    -1

#endif /* _NEXUS_INCLUDE_NEXUS_H */
