/*
 * types.h
 *
 * rcsid = "/net/bcomp/src/master/nexus/types.h,v 1.38 1995/10/23 05:27:03 tuecke Exp"
 */

#ifndef _NEXUS_INCLUDE_TYPES_H
#define _NEXUS_INCLUDE_TYPES_H


#ifdef NEXUS_PROFILE

#ifndef NEXUS_PROFILE_RSR_HASH_TABLE_SIZE
#define NEXUS_PROFILE_RSR_HASH_TABLE_SIZE	16
#endif

/*
 * nexus_profile_rsr_record_t
 *
 * This is used to keep cumulative totals of handler invocations
 */
typedef struct _nexus_profile_rsr_record_t
{
    int						node_id;
    int						context_id;
    int						count;
    int						size;
    struct _nexus_profile_rsr_record_t *	next;
} nexus_profile_rsr_record_t;

#endif /* NEXUS_PROFILE */


/*
 * nexus_handler_record_t
 *
 * One record in the handler table...
 */
typedef struct _nexus_handler_record_t
{
    char *				name;
    int					id;
    nexus_handler_type_t		type;
    nexus_handler_func_t		func;
    struct _nexus_handler_record_t *	next;
#ifdef NEXUS_PROFILE
    nexus_profile_rsr_record_t **	profile_table;
#endif    
} nexus_handler_record_t;


/*
 * nexus_segment_t
 *
 * Holds one data segment in a context.
 */
typedef struct _nexus_segment_t
{
    void *			data;
    unsigned int		size;
    struct _nexus_context_t *	context;
    struct _nexus_segment_t *	prev;
    struct _nexus_segment_t *	next;
} nexus_segment_t;

#ifdef NEXUS_ALIGN
#define NEXUS_SEGMENT_HEADER_SIZE \
    (sizeof(nexus_segment_t) \
     + ((NEXUS_ALIGN - (sizeof(nexus_segment_t) % NEXUS_ALIGN)) % NEXUS_ALIGN))
#else  /* NEXUS_ALIGN */
#define NEXUS_SEGMENT_HEADER_SIZE \
    sizeof(nexus_segment_t)
#endif /* NEXUS_ALIGN */


/*
 * nexus_context_t
 *
 * Hold context info.
 *
 * The segment_list is a circular, doubly linked list, with a dummy
 * first segment.
 */
typedef struct _nexus_context_t
{
    nexus_mutex_t		mutex;
    int				id;
    void *                      handle;
    int                         (*entry_point)(void *);
    void                        (*switchingfunc)(void *);
    int                         (*destructofunc)(struct _nexus_context_t *);
    int                         (*pfnNexusBoot)(void);
    void                        (*pfnNexusExit)(void);
    void                        (*pfnNexusUnknownHandler)(void *,
							  nexus_buffer_t *,
							  char *, int );
    int                         (*pfnNexusAcquiredAsNode)(void);
    int				n_segments;
    struct _nexus_segment_t *	segment_list;
    void *			initial_segment;
    nexus_handler_record_t **	handler_table;
    nexus_bool_t		is_node;
    struct _nexus_context_t *	context_list_next;
#ifdef NEXUS_PROFILE
    int				snapshot_id;
    int				rsr_profile_count;
    nexus_global_pointer_t	master_gp;
    nexus_bool_t		waiting_on_master_gp;
    nexus_cond_t		cond;
#endif /* NEXUS_PROFILE */
} nexus_context_t;


/*
 * nexus_new_nodes_monitor_t
 *
 * This monitor is used to keep track of new processes as they checkin.
 */
typedef struct _nexus_new_nodes_monitor_t
{
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
    int			total;
    int			checked_in;
    nexus_node_t *	nodes;
} nexus_new_nodes_monitor_t;


typedef struct _nexus_startup_node_t nexus_startup_node_t;
typedef struct _nexus_startup_funcs_t nexus_startup_funcs_t;

/*
 * nexus_startup_funcs_t
 *
 * This is the function indirection table used by startup modules.
 */
struct _nexus_startup_funcs_t
{
    nexus_bool_t	(*preinit)(nexus_startup_node_t **nl_head,
				   nexus_startup_node_t **nl_tail);
    char *		(*get_master_node_name)(void);
    void		(*init)(int *argc, char ***argv);
    void		(*shutdown)(nexus_bool_t shutdown_others);
    void		(*abort)(void);
    int    		(*start_node)(nexus_startup_node_t *node,
				      nexus_global_pointer_t *reply_gp,
				      int first_checkin_number);
    nexus_bool_t	(*start_context)(char *executable_path,
					 int context_size,
					 nexus_global_pointer_t *reply_gp,
					 int checkin_number,
					 int *return_code,
					 nexus_context_t **new_local_context);
    void                (*register_handlers)(void);
};


/*
 * nexus_startup_node_t.
 *
 * Startup nodes structure.  st_iface.c creates this structure and
 * passes it to the start_node function of the startup module, to
 * tell it what start a node or set of nodes.
 *
 * If 'count' is > 1, then 'number' and 'id' fields are the _first_
 * of a continguous set of numbers and ids that the set of nodes should use.  
 */
struct _nexus_startup_node_t
{
    char *			name;
    int				number;
    int				count;
    int                         count_other_nodes;
    int				id;
    char *			directory_path;
    char *			executable_path;
    char *			startup_module;
    nexus_startup_funcs_t *	startup_funcs;
    nexus_startup_node_t *	next;
    void *			st_ptr;
};


typedef struct _nexus_database_funcs_t nexus_database_funcs_t;

/* 
 * This is the function indirection table that gets loaded in
 * db_iface
 */
struct _nexus_database_funcs_t
{
    nexus_bool_t                (*init)(int *argc, char ***argv);
    void                        (*usage)(void);
    int                         (*new_params)(char *buf,
					     int size);
    char *                      (*lookup)(char *node_name,
				 	 int node_number,
					 char *key);
    void                        (*shutdown)(void);
    void                        (*abort)(void);
};

typedef struct _nexus_list_t nexus_list_t;

/*
 * This is a generic linked list
 */
struct _nexus_list_t
{
    void *                    value;
    struct _nexus_list_t *    next;
};

typedef struct _nexus_db_hash_entry_t nexus_db_hash_entry_t;

/*
 * This is the hash table stuff for database/startup file lookup
 * values.
 */
struct _nexus_db_hash_entry_t
{
    char *name;
    int number;
    int count;
    nexus_list_t *attr;
    struct _nexus_db_hash_entry_t *next;
};

#endif /* _NEXUS_INCLUDE_TYPES_H */
